@@ -1,54 +1,120 @@
-from collections import Iterable
+from collections import OrderedDict
+from decimal import Decimal
 from difflib import unified_diff
+from functools import partial as partial_type, partial
 from pprint import pformat
 from re import compile, MULTILINE
 from types import GeneratorType
 
 from testfixtures import not_there
-from testfixtures.compat import ClassType, Unicode, basestring, PY3
+from testfixtures.compat import ClassType, Iterable, Unicode, basestring, PY3, PY2
 from testfixtures.resolve import resolve
 from testfixtures.utils import indent
 from testfixtures.mock import parent_name, mock_call, unittest_mock_call
 
 
+def diff(x, y, x_label='', y_label=''):
+    """
+    A shorthand function that uses :mod:`difflib` to return a
+    string representing the differences between the two string
+    arguments.
+
+    Most useful when comparing multi-line strings.
+    """
+    return '\n'.join(
+        unified_diff(
+            x.split('\n'),
+            y.split('\n'),
+            x_label or 'first',
+            y_label or 'second',
+            lineterm='')
+    )
+
+
 def compare_simple(x, y, context):
     """
     Returns a very simple textual difference between the two supplied objects.
     """
     if x != y:
-        return context.label('x', repr(x)) + ' != ' + context.label('y', repr(y))
+        repr_x = repr(x)
+        repr_y = repr(y)
+        if repr_x == repr_y:
+            if type(x) is not type(y):
+                return compare_with_type(x, y, context)
+            x_attrs = _extract_attrs(x)
+            y_attrs = _extract_attrs(y)
+            diff = _compare_mapping(x_attrs, y_attrs, context, x,
+                                    'attributes ', '.%s')
+            if not diff:
+                return 'Both %s and %s appear as %r, but are not equal!' % (
+                    context.x_label or 'x', context.y_label or 'y', repr_x
+                )
+            return diff
+        return context.label('x', repr_x) + ' != ' + context.label('y', repr_y)
 
 
-def _extract_attrs(obj):
-    slots = getattr(obj, '__slots__', None)
-    if slots:
-        attrs = {}
+def _extract_attrs(obj, ignore=None):
+    try:
+        attrs = vars(obj).copy()
+    except TypeError:
+        attrs = None
+    else:
+        if isinstance(obj, BaseException):
+            attrs['args'] = obj.args
+
+    has_slots = getattr(obj, '__slots__', not_there) is not not_there
+    if has_slots:
+        slots = set()
+        for cls in type(obj).__mro__:
+            slots.update(getattr(cls, '__slots__', ()))
+        if slots and attrs is None:
+            attrs = {}
         for n in slots:
             value = getattr(obj, n, not_there)
             if value is not not_there:
                 attrs[n] = value
-    else:
-        try:
-            attrs = vars(obj).copy()
-        except TypeError:
-            return None
-        else:
-            if isinstance(obj, BaseException):
-                attrs['args'] = obj.args
+
+    if attrs is None:
+        return None
+
+    if ignore is not None:
+        for attr in ignore:
+            attrs.pop(attr, None)
     return attrs
 
 
-def compare_object(x, y, context):
+def _attrs_to_ignore(context, ignore_attributes, obj):
+    ignore = context.get_option('ignore_attributes', ())
+    if isinstance(ignore, dict):
+        ignore = ignore.get(type(obj), ())
+    ignore = set(ignore)
+    ignore.update(ignore_attributes)
+    return ignore
+
+
+def compare_object(x, y, context, ignore_attributes=()):
     """
     Compare the two supplied objects based on their type and attributes.
+
+    :param ignore_attributes:
+
+       Either a sequence of strings containing attribute names to be ignored
+       when comparing or a mapping of type to sequence of strings containing
+       attribute names to be ignored when comparing that type.
+
+       This may be specified as either a parameter to this function or in the
+       ``context``. If specified in both, they will both apply with precedence
+       given to whatever is specified is specified as a parameter.
+       If specified as a parameter to this fucntion, it may only be a list of
+       strings.
     """
-    if type(x) is not type(y) or isinstance(x, ClassType):
+    if type(x) is not type(y) or isinstance(x, (ClassType, type)):
         return compare_simple(x, y, context)
-    x_attrs = _extract_attrs(x)
-    y_attrs = _extract_attrs(y)
-    if x_attrs is None or y_attrs is None:
+    x_attrs = _extract_attrs(x, _attrs_to_ignore(context, ignore_attributes, x))
+    y_attrs = _extract_attrs(y, _attrs_to_ignore(context, ignore_attributes, y))
+    if x_attrs is None or y_attrs is None or not (x_attrs and y_attrs):
         return compare_simple(x, y, context)
-    if x_attrs != y_attrs:
+    if context.ignore_eq or x_attrs != y_attrs:
         return _compare_mapping(x_attrs, y_attrs, context, x,
                                 'attributes ', '.%s')
 
@@ -79,7 +145,7 @@ def compare_with_type(x, y, context):
     return '{x} != {y}'.format(**to_render)
 
 
-def compare_sequence(x, y, context):
+def compare_sequence(x, y, context, prefix=True):
     """
     Returns a textual description of the differences between the two
     supplied sequences.
@@ -95,7 +161,7 @@ def compare_sequence(x, y, context):
     if l_x == l_y and i == l_x:
         return
 
-    return ('sequence not as expected:\n\n'
+    return (('sequence not as expected:\n\n' if prefix else '')+
             'same:\n%s\n\n'
             '%s:\n%s\n\n'
             '%s:\n%s') % (pformat(x[:i]),
@@ -314,11 +380,42 @@ def compare_bytes(x, y, context):
 def compare_call(x, y, context):
     if x == y:
         return
-    x_name, x_args, x_kw = x
-    y_name, y_args, y_kw = y
+
+    def extract(call):
+        try:
+            name, args, kwargs = call
+        except ValueError:
+            name = None
+            args, kwargs = call
+        return name, args, kwargs
+
+    x_name, x_args, x_kw = extract(x)
+    y_name, y_args, y_kw = extract(y)
+
     if x_name == y_name and x_args == y_args and x_kw == y_kw:
         return compare_call(getattr(x, parent_name), getattr(y, parent_name), context)
-    return compare_text(repr(x), repr(y), context)
+
+    if repr(x) != repr(y):
+        return compare_text(repr(x), repr(y), context)
+
+    different = (
+        context.different(x_name, y_name, ' function name') or
+        context.different(x_args, y_args, ' args') or
+        context.different(x_kw, y_kw, ' kw')
+    )
+    if not different:
+        return
+
+    return 'mock.call not as expected:'
+
+
+
+def compare_partial(x, y, context):
+    x_attrs = dict(func=x.func, args=x.args, keywords=x.keywords)
+    y_attrs = dict(func=y.func, args=y.args, keywords=y.keywords)
+    if x_attrs != y_attrs:
+        return _compare_mapping(x_attrs, y_attrs, context, x,
+                                'attributes ', '.%s')
 
 
 def _short_repr(obj):
@@ -337,10 +434,12 @@ _registry = {
     Unicode: compare_text,
     int: compare_simple,
     float: compare_simple,
+    Decimal: compare_simple,
     GeneratorType: compare_generator,
     mock_call.__class__: compare_call,
     unittest_mock_call.__class__: compare_call,
     BaseException: compare_exception,
+    partial_type: compare_partial,
     }
 
 if PY3:
@@ -414,6 +513,13 @@ class CompareContext(object):
         if actual is not not_there:
             possible.append(actual)
 
+        x = self.options.pop('x', not_there)
+        if x is not not_there:
+            possible.append(x)
+        y = self.options.pop('y', not_there)
+        if y is not not_there:
+            possible.append(y)
+
         if len(possible) != 2:
             message = 'Exactly two objects needed, you supplied:'
             if possible:
@@ -460,6 +566,9 @@ class CompareContext(object):
         return '\n\nWhile comparing %s: ' % ''.join(self.breadcrumbs[1:])
 
     def seen(self, x, y):
+        # don't get confused by string interning:
+        if isinstance(x, basestring) and isinstance(y, basestring):
+            return False
         key = id(x), id(y)
         if key in self._seen:
             return True
@@ -511,14 +620,17 @@ class CompareContext(object):
 
 def compare(*args, **kw):
     """
-    Compare the two arguments passed either positionally or using
-    explicit ``expected`` and ``actual`` keyword paramaters. An
-    :class:`AssertionError` will be raised if they are not the same.
-    The :class:`AssertionError` raised will attempt to provide
+    Compare two objects, raising an :class:`AssertionError` if they are not
+    the same. The :class:`AssertionError` raised will attempt to provide
     descriptions of the differences found.
 
+    The two objects to compare can be passed either positionally or using
+    explicit keyword arguments named ``x`` and ``y``, or ``expected`` and
+    ``actual``.
+
     Any other keyword parameters supplied will be passed to the functions
-    that end up doing the comparison. See the API documentation below
+    that end up doing the comparison. See the
+    :mod:`API documentation below <testfixtures.comparison>`
     for details of these.
 
     :param prefix: If provided, in the event of an :class:`AssertionError`
@@ -529,6 +641,18 @@ def compare(*args, **kw):
                    being raised, the suffix supplied will be appended to the
                    message in the :class:`AssertionError`.
 
+    :param x_label: If provided, in the event of an :class:`AssertionError`
+                    being raised, the object passed as the first positional
+                    argument, or ``x`` keyword argument, will be labelled
+                    with this string in the message in the
+                    :class:`AssertionError`.
+
+    :param y_label: If provided, in the event of an :class:`AssertionError`
+                    being raised, the object passed as the second positional
+                    argument, or ``y`` keyword argument, will be labelled
+                    with this string in the message in the
+                    :class:`AssertionError`.
+
     :param raises: If ``False``, the message that would be raised in the
                    :class:`AssertionError` will be returned instead of the
                    exception being raised.
@@ -575,32 +699,69 @@ def compare(*args, **kw):
     return message
 
 
-class Comparison(object):
+class StatefulComparison(object):
     """
-    These are used when you need to compare objects
-    that do not natively support comparison.
+    A base class for stateful comparison objects.
+    """
+
+    failed = ''
+    expected = None
+    name_attrs = ()
+
+    def __eq__(self, other):
+        return not(self != other)
+
+    def name(self):
+        name = type(self).__name__
+        if self.name_attrs:
+            name += '(%s)' % ', '.join('%s=%r' % (n, getattr(self, n)) for n in self.name_attrs)
+        return name
+
+    def body(self):
+        return pformat(self.expected)[1:-1]
+
+    def __repr__(self):
+        name = self.name()
+        body = self.failed or self.body()
+        prefix = '<%s%s>' % (name, self.failed and '(failed)' or '')
+        if '\n' in body:
+            return '\n'+prefix+'\n'+body.strip('\n')+'\n'+'</%s>' % name
+        elif body:
+            return prefix + body + '</>'
+        return prefix
+
+
+class Comparison(StatefulComparison):
+    """
+    These are used when you need to compare an object's type, a subset of its attributes
+    or make equality checks with objects that do not natively support comparison.
 
     :param object_or_type: The object or class from which to create the
                            :class:`Comparison`.
 
-    :param attribute_dict: An optional dictionary containing attibutes
+    :param attribute_dict: An optional dictionary containing attributes
                            to place on the :class:`Comparison`.
 
-    :param strict: If true, any expected attributes not present or extra
-                   attributes not expected on the object involved in the
-                   comparison will cause the comparison to fail.
+    :param partial:
+      If true, only the specified attributes will be checked and any extra attributes
+      of the object being compared with will be ignored.
 
     :param attributes: Any other keyword parameters passed will placed
                        as attributes on the :class:`Comparison`.
-    """
 
-    failed = None
+    :param strict:
+
+      .. deprecated:: 6.16.0
+
+        Use ``partial`` instead.
+    """
 
     def __init__(self,
                  object_or_type,
                  attribute_dict=None,
-                 strict=True,
+                 partial=False,
                  **attributes):
+        self.partial = partial or not attributes.pop('strict', True)
         if attributes:
             if attribute_dict is None:
                 attribute_dict = attributes
@@ -618,73 +779,302 @@ class Comparison(object):
             c = object_or_type.__class__
             if attribute_dict is None:
                 attribute_dict = _extract_attrs(object_or_type)
-        self.c = c
-        self.v = attribute_dict
-        self.strict = strict
+        self.expected_type = c
+        self.expected_attributes = attribute_dict
 
-    def __eq__(self, other):
-        if self.c is not other.__class__:
+    def __ne__(self, other):
+        # .__class__ is important for Py2 compatibility.
+        if self.expected_type is not other.__class__:
             self.failed = 'wrong type'
-            return False
-
-        if self.v is None:
             return True
 
-        if self.strict:
-            v = _extract_attrs(other)
+        if self.expected_attributes is None:
+            return False
+
+        attribute_names = set(self.expected_attributes.keys())
+        if self.partial:
+            actual_attributes = {}
         else:
-            v = {}
-            for k in self.v.keys():
-                try:
-                    v[k] = getattr(other, k)
-                except AttributeError:
-                    pass
+            actual_attributes = _extract_attrs(other)
+            attribute_names -= set(actual_attributes)
+
+        for name in attribute_names:
+            try:
+                actual_attributes[name] = getattr(other, name)
+            except AttributeError:
+                pass
 
         kw = {'x_label': 'Comparison', 'y_label': 'actual'}
         context = CompareContext(kw)
-        self.failed = _compare_mapping(self.v,
-                                       v,
+        self.failed = _compare_mapping(self.expected_attributes,
+                                       actual_attributes,
                                        context,
                                        obj_for_class=not_there,
                                        prefix='attributes ',
                                        breadcrumb='.%s',
-                                       check_y_not_x=self.strict)
-        return not self.failed
+                                       check_y_not_x=not self.partial)
+        return bool(self.failed)
 
-    def __ne__(self, other):
-        return not(self == other)
+    def name(self):
+        name = 'C:'
+        module = getattr(self.expected_type, '__module__', None)
+        if module:
+            name = name + module + '.'
+        name += (getattr(self.expected_type, '__name__', None) or repr(self.expected_type))
+        return name
 
-    def __repr__(self):
-        name = getattr(self.c, '__module__', '')
-        if name:
-            name += '.'
-        name += getattr(self.c, '__name__', '')
-        if not name:
-            name = repr(self.c)
-
-        text = ''
-        if self.failed:
-            text = self.failed
-        elif self.v:
+    def body(self):
+        if self.expected_attributes:
+            # if we're not failed, show what we will expect:
             lines = []
-            for k, v in sorted(self.v.items()):
+            for k, v in sorted(self.expected_attributes.items()):
                 rv = repr(v)
                 if '\n' in rv:
                     rv = indent(rv)
                 lines.append('%s: %s' % (k, rv))
-            text = '\n'.join(lines)
-            if len(lines) > 1:
-                text = '\n'+text
+            return '\n'.join(lines)
+        return ''
 
-        r = '<C%s:%s>' % (self.failed and '(failed)' or '', name)
-        if '\n' in text:
-            r = '\n'+r+text+'\n'
+
+class SequenceComparison(StatefulComparison):
+    """
+    An object that can be used in comparisons of expected and actual
+    sequences.
+
+    :param expected: The items expected to be in the sequence.
+    :param ordered:
+      If the items are expected to be in the order specified.
+      Defaults to ``True``.
+    :param partial:
+      If any items not expected should be ignored.
+      Defaults to ``False``.
+    :param recursive:
+      If a difference is found, recursively compare the item where
+      the difference was found to highlight exactly what was different.
+      Defaults to ``False``.
+    """
+
+    name_attrs = ('ordered', 'partial')
+
+    def __init__(self, *expected, **kw):
+        self.expected = expected
+        # py2 :-(
+        self.ordered = kw.pop('ordered', True)
+        self.partial = kw.pop('partial', False)
+        self.recursive = kw.pop('recursive', False)
+        assert not kw, 'unexpected parameter'
+        self.checked_indices = set()
+
+    def __ne__(self, other):
+        try:
+            actual = original_actual = list(other)
+        except TypeError:
+            self.failed = 'bad type'
+            return True
+        expected = list(self.expected)
+        actual = list(actual)
+
+        matched = []
+        matched_expected_indices = []
+        matched_actual_indices = []
+
+        missing_from_expected = actual
+        missing_from_expected_indices = actual_indices = list(range(len(actual)))
+
+        missing_from_actual = []
+        missing_from_actual_indices = []
+
+        for e_i, e in enumerate(expected):
+            try:
+                i = actual.index(e)
+                a_i = actual_indices.pop(i)
+            except ValueError:
+                missing_from_actual.append(e)
+                missing_from_actual_indices.append(e_i)
+            else:
+                matched.append(missing_from_expected.pop(i))
+                matched_expected_indices.append(e_i)
+                matched_actual_indices.append(a_i)
+                self.checked_indices.add(a_i)
+
+        matches_in_order = matched_actual_indices == sorted(matched_actual_indices)
+        all_matched = not (missing_from_actual or missing_from_expected)
+        partial_match = self.partial and not missing_from_actual
+
+        if (matches_in_order or not self.ordered) and (all_matched or partial_match):
+            return False
+
+        expected_indices = matched_expected_indices+missing_from_actual_indices
+        actual_indices = matched_actual_indices
+
+        if self.partial:
+            # try to give a clue as to what didn't match:
+            if self.recursive and self.ordered and missing_from_expected:
+                actual_indices.append(missing_from_expected_indices.pop(0))
+                missing_from_expected.pop(0)
+
+            ignored = missing_from_expected
+            missing_from_expected = None
         else:
-            r += text
-        if text:
-            r += '</C>'
+            actual_indices += missing_from_expected_indices
+            ignored = None
 
-        return r
+        message = []
+
+        def add_section(name, content):
+            if content:
+                message.append(name+':\n'+pformat(content))
+
+        add_section('ignored', ignored)
+
+        if self.ordered:
+            message.append(compare(
+                expected=[self.expected[i] for i in sorted(expected_indices)],
+                actual=[original_actual[i] for i in sorted(actual_indices)],
+                recursive=self.recursive,
+                raises=False
+            ).split('\n\n', 1)[1])
+        else:
+            add_section('same', matched)
+            add_section('in expected but not actual', missing_from_actual)
+            add_section('in actual but not expected', missing_from_expected)
+
+        self.failed = '\n\n'.join(message)
+        return True
+
+
+class Subset(SequenceComparison):
+    """
+    A shortcut for :class:`SequenceComparison` that checks if the
+    specified items are present in the sequence.
+    """
+
+    name_attrs = ()
+
+    def __init__(self, *expected):
+        super(Subset, self).__init__(*expected, ordered=False, partial=True)
+
+
+class Permutation(SequenceComparison):
+    """
+    A shortcut for :class:`SequenceComparison` that checks if the set of items
+    in the sequence is as expected, but without checking ordering.
+    """
+
+    def __init__(self, *expected):
+        super(Permutation, self).__init__(*expected, ordered=False, partial=False)
+
+
+class MappingComparison(StatefulComparison):
+    """
+    An object that can be used in comparisons of expected and actual
+    mappings.
+
+    :param expected_mapping:
+      The mapping that should be matched expressed as either a sequence of
+      ``(key, value)`` tuples or a mapping.
+    :param expected_items: The items that should be matched.
+    :param ordered:
+      If the keys in the mapping are expected to be in the order specified.
+      Defaults to ``False``.
+    :param partial:
+      If any keys not expected should be ignored.
+      Defaults to ``False``.
+    :param recursive:
+      If a difference is found, recursively compare the value where
+      the difference was found to highlight exactly what was different.
+      Defaults to ``False``.
+    """
+
+    name_attrs = ('ordered', 'partial')
+
+    def __init__(self, *expected_mapping, **expected_items):
+        # py2 :-(
+        self.ordered = expected_items.pop('ordered', False)
+        self.partial = expected_items.pop('partial', False)
+        self.recursive = expected_items.pop('recursive', False)
+
+        if PY2 and self.ordered:
+            if expected_items:
+                raise TypeError('order undefined on Python 2')
+            elif expected_mapping and type(expected_mapping[0]) is dict:
+                raise TypeError('dict order undefined on Python 2')
+
+        if len(expected_mapping) == 1:
+            expected = OrderedDict(*expected_mapping)
+        else:
+            expected = OrderedDict(expected_mapping)
+        expected.update(expected_items)
+
+        self.expected = expected
+
+    def body(self):
+        # this can all go away and use the super class once py2 is gone :'(
+        parts = []
+        text_length = 0
+        for key, value in self.expected.items():
+            part = repr(key)+': '+pformat(value)
+            text_length += len(part)
+            parts.append(part)
+        if text_length > 60:
+            sep = ',\n'
+        else:
+            sep = ', '
+        return sep.join(parts)
+
+    def __ne__(self, other):
+        try:
+            actual_keys = other.keys()
+            actual_mapping = dict(other.items())
+        except AttributeError:
+            self.failed = 'bad type'
+            return True
+
+        expected_keys = self.expected.keys()
+        expected_mapping = self.expected
+
+        if self.partial:
+            ignored_keys = set(actual_keys) - set(expected_keys)
+            for key in ignored_keys:
+                del actual_mapping[key]
+            # preserve the order:
+            actual_keys = [k for k in actual_keys if k not in ignored_keys]
+        else:
+            ignored_keys = None
+
+        mapping_differences = compare(
+            expected=expected_mapping,
+            actual=actual_mapping,
+            recursive=self.recursive,
+            raises=False
+        )
+
+        if self.ordered:
+            key_differences = compare(
+                expected=list(expected_keys),
+                actual=list(actual_keys),
+                recursive=self.recursive,
+                raises=False
+            )
+        else:
+            key_differences = None
+
+        if key_differences or mapping_differences:
+
+            message = []
+
+            if ignored_keys:
+                message.append('ignored:\n'+pformat(sorted(ignored_keys)))
+
+            if mapping_differences:
+                message.append(mapping_differences.split('\n\n', 1)[1])
+
+            if key_differences:
+                message.append('wrong key order:\n\n'+key_differences.split('\n\n', 1)[1])
+
+            self.failed = '\n\n'.join(message)
+
+            return True
 
 
 class StringComparison:
@@ -749,24 +1139,6 @@ class RoundComparison:
         return '<R:%s to %i digits>' % (self.rounded, self.precision)
 
 
-def diff(x, y, x_label='', y_label=''):
-    """
-    A shorthand function that uses :mod:`difflib` to return a
-    string representing the differences between the two string
-    arguments.
-
-    Most useful when comparing multi-line strings.
-    """
-    return '\n'.join(
-        unified_diff(
-            x.split('\n'),
-            y.split('\n'),
-            x_label or 'first',
-            y_label or 'second',
-            lineterm='')
-    )
-
-
 class RangeComparison:
     """
     An object that can be used in comparisons of orderable types to
